unit VideoPlayer;

interface

uses
  Windows, ActiveX, MMSystem, DShow;

const
  IID_IMediaDet: TGUID = '{65BD0710-24D2-4FF7-9324-ED2E5D3ABAFA}';
  CLSID_MediaDet: TGUID = '{65BD0711-24D2-4FF7-9324-ED2E5D3ABAFA}';
  IID_IBasicVideo2: TGUID = (D1:$329bb360;D2:$f6ea;D3:$11d1;D4:($90,$38,$00,$a0,$c9,$69,$72,$98));
  IID_ISpecifyPropertyPages: TGUID = '{B196B28B-BAB4-101A-B69C-00AA00341D07}';

type
  // DShow.pasɂȂC^[tF[X
  ISampleGrabberCB = interface(IUnknown)
    ['{0579154A-2B53-4994-B0D0-E773148EFF85}']
    function SampleCB(SampleTime: Double; pSample: IMediaSample): HResult; stdcall;
    function BufferCB(SampleTime: Double; pBuffer: PByte; BufferLen: longint): HResult; stdcall;
  end;
  ISampleGrabber = interface(IUnknown)
    ['{6B652FFF-11FE-4FCE-92AD-0266B5D7C78F}']
    function SetOneShot(OneShot: BOOL): HResult; stdcall;
    function SetMediaType(var pType: TAM_MEDIA_TYPE): HResult; stdcall;
    function GetConnectedMediaType(out pType: TAM_MEDIA_TYPE): HResult; stdcall;
    function SetBufferSamples(BufferThem: BOOL): HResult; stdcall;
    function GetCurrentBuffer(var pBufferSize: longint; pBuffer: Pointer): HResult; stdcall;
    function GetCurrentSample(out ppSample: IMediaSample): HResult; stdcall;
    function SetCallback(pCallback: ISampleGrabberCB; WhichMethodToCallback: longint): HResult; stdcall;
  end;
  IMediaDet = interface(IUnknown)
    ['{65BD0710-24D2-4FF7-9324-ED2E5D3ABAFA}']
    function get_Filter(out pVal: IUnknown): HResult; stdcall;
    function put_Filter(newVal: IUnknown): HResult; stdcall;
    function get_OutputStreams(out pVal: longint): HResult; stdcall;
    function get_CurrentStream(out pVal: longint): HResult; stdcall;
    function put_CurrentStream(newVal: longint): HResult; stdcall;
    function get_StreamType(out pVal: TGUID): HResult; stdcall;
    function get_StreamTypeB(out pVal: WideString): HResult; stdcall;
    function get_StreamLength(out pVal: Double): HResult; stdcall;
    function get_Filename(out pVal: WideString): HResult; stdcall;
    function put_Filename(pVal: WideString): HResult; stdcall;
    function GetBitmapBits(streamTime: Double; var pBufferSize: longint; var pBuffer: Pchar;
      Width: longint; Height: longint): HResult; stdcall;
    function WriteBitmapBits(streamTime: Double; Width: longint; Height: longint;
      Filename: WideString): HResult; stdcall;
    function get_StreamMediaType(out pVal: TAM_MEDIA_TYPE): HResult; stdcall;
    function GetSampleGrabber(out ppVal: ISampleGrabber): HResult; stdcall;
    function get_FrameRate(out pVal: Double): HResult; stdcall;
    function EnterBitmapGrabMode(SeekTime: Double): HResult; stdcall;
  end;
  IBasicVideo2 = interface(IBasicVideo)
    ['{329bb360-f6ea-11d1-9038-00a0c9697298}']
    function GetPreferredAspectRatio(out plAspectX, plAspectY: Longint): HRESULT; stdcall;
  end;

  // DShow.pas̒`܂Ă
  //   wPriority: DWORD -> WORD
  //   wLanguage: DWORD -> WORD
  TAVIStreamHeader2 = packed record
    fcc: FOURCC;            // 'strh'
    cb: DWORD;              // size of this structure - 8

    fccType: FOURCC;        // stream type codes

    fccHandler: FOURCC;
    dwFlags: DWORD;

    wPriority: WORD;
    wLanguage: WORD;
    dwInitialFrames: DWORD;
    dwScale: DWORD;
    dwRate: DWORD;          // dwRate/dwScale is stream tick rate in ticks/sec
    dwStart: DWORD;
    dwLength: DWORD;
    dwSuggestedBufferSize: DWORD;
    dwQuality: DWORD;
    dwSampleSize: DWORD;

    rcFrame: packed record
      left: SmallInt;
      top: SmallInt;
      right: SmallInt;
      bottom: SmallInt;
    end;
 end;

  TVideoFileInfo = record
    Width: Integer;
    Height: Integer;
    Length: Integer; // msec
    FramePerSec: Double;
    Aspect: Double;
    AVIMainHeader: TAVIMainHeader;
    BitmapInfoHeader: TBitmapInfoHeader;
    WaveFormat: TWaveFormat;
  end;

  TVPState = (vpPlay, vpPause, vpStop);

  TVideoPlayer = class(TObject)
  protected
    FParent: HWND;
    FBuilder: IGraphBuilder;
    FControl: IMediaControl;
    FSeeking: IMediaSeeking;
    FEvent: IMediaEvent;
    FBasicVideo: IBasicVideo;
    FBasicVideo2: IBasicVideo2;
    //Fivf: IVideoFrameStep;
    FAudio: IBasicAudio;
    FWindow: IVideoWindow;
    FTimerID: Integer;
    FDuration: LONGLONG;
    FState: TVpState;
    FLoaded: Boolean;
    FFileInfo: TVideoFileInfo;
    FFileName: string;
    procedure StartTimer;
    procedure StopTimer;
    function GetVolume: Integer;
    procedure SetVolume(Value: Integer);
    function GetPosition: Integer;
    procedure SetPosition(Value: Integer);
    function SearchChunk(var F: File; Id: PChar; var Size: Integer): Boolean;
    //procedure GetMpegInfo(var F: File; var Aspect: Double; var w, h: Integer);
    function GetAVIHeader(const FileName: string): Boolean;
  public
    FilterNames: array[0..255] of string;
    FilterCount: Integer;
    OnTimer: procedure of object;
    OnEndPlay: procedure of object;
    constructor Create(Parent: HWND); virtual;
    procedure FreeInterface;
    function LoadFile(const FileName: string): HResult;
    procedure Play;
    procedure Stop;
    procedure Pause;
    procedure PauseOnly;
    procedure SetWindowPosition(ALeft, ATop, AWidth, AHeight: Longint);
    function WriteBitmapBits(const FileName: string): HRESULT;
    procedure GetFilterNames;
    function ShowFilterPropertyByName(const FilterName: string): HRESULT;
    property FileName: string read FFileName;
    property Loaded: Boolean read FLoaded;
    property Volume: Integer read GetVolume write SetVolume; // -10000..0
    property Position: Integer read GetPosition write SetPosition; // msec
    property State: TVpState read FState;
    property FileInfo: TVideoFileInfo read FFileInfo;
  end;

  function GetDShowErrMsg(Code: HRESULT): string;

implementation

(*
const
  Mpeg1AspectTable: array[0..14] of Double = (
    1, 1, 0.6735, 0.7031, 0.7615, 0.8055, 0.8437, 0.8935,
    0.9815, 54/59 {PAL}, 1.0255, 1.0695, 11/10 {NTSC}, 1.1575, 1.2015
  );
*)

function CompareTGUID(const A, B: TGUID): Boolean;
begin
  Result := (A.D1 = B.D1) and (A.D2 = B.D2) and (A.D3 = B.D3) and
    (A.D4[0] = B.D4[0]) and (A.D4[1] = B.D4[1]) and (A.D4[2] = B.D4[2]) and
    (A.D4[3] = B.D4[3]) and (A.D4[4] = B.D4[4]) and (A.D4[5] = B.D4[5]) and
    (A.D4[6] = B.D4[6]) and (A.D4[7] = B.D4[7]);
end;

function GetDShowErrMsg(Code: HRESULT): string;
var
  Buf: array[0..MAX_ERROR_TEXT_LEN-1] of Char;
begin
  SetString(Result, Buf, AMGetErrorText(Code, Buf, SizeOf(Buf)));
end;

procedure TimeProc(IDEvent, uReserved, dwUser, dwReserved1, dwReserved2: Integer); stdcall;
begin
  with TVideoPlayer(dwUser) do
  begin
    if FLoaded and Assigned(OnTimer) then
      OnTimer;
  end;
end;

constructor TVideoPlayer.Create(Parent: HWND);
begin
  FParent := Parent;
end;

procedure TVideoPlayer.StartTimer;
begin
  StopTimer;
  FTimerID := timeSetEvent(100, 50, TimeProc, DWORD(Self), TIME_PERIODIC);
end;

procedure TVideoPlayer.StopTimer;
begin
  if FTimerID <> 0 then
  begin
    timeKillEvent(FTimerID);
    FTimerID := 0;
  end;
end;

function TVideoPlayer.GetVolume: Integer;
begin
  if FLoaded then
    FAudio.get_Volume(Result)
  else
    Result := 0;
end;

procedure TVideoPlayer.SetVolume(Value: Integer);
begin
  if FLoaded then
    FAudio.put_Volume(Value);
end;

function TVideoPlayer.GetPosition: Integer;
var
  cp: LONGLONG;
begin
  if FLoaded then
  begin
    FSeeking.GetCurrentPosition(cp);
    if cp >= FDuration then
    begin
      FControl.Stop; // ̂ȁ[
      FState := vpStop;
      if Assigned(OnEndPlay) then
        OnEndPlay;
    end;
    Result := Trunc(cp / 10000);
  end else
    Result := 0;
end;

procedure TVideoPlayer.SetPosition(Value: Integer);
const
  a: LONGLONG = 10000;
var
  CurPos, StopPos: LONGLONG;
begin
  if FLoaded then
  begin
    CurPos := Value * a;
    if CurPos < 0 then
      CurPos := 0
    else if CurPos >= FDuration then
      CurPos := FDuration;
    StopPos := 0;
    FSeeking.SetPositions(CurPos, AM_SEEKING_AbsolutePositioning,
      StopPos, AM_SEEKING_NoPositioning);
    GetPosition;
  end;
end;

procedure TVideoPlayer.FreeInterface;
begin
  if Assigned(FControl) then
    FControl.Stop;
  if Assigned(FWindow) then
  begin
    FWindow.put_Visible(False);
    FWindow.put_Owner(OAHWND(0));
  end;
  FWindow := nil;
  FAudio := nil;
  //Fivf := nil;
  FBasicVideo2 := nil;
  FBasicVideo := nil;
  FEvent := nil;
  FSeeking := nil;
  FControl := nil;
  FBuilder := nil;
  FLoaded := False;
end;

procedure TVideoPlayer.Play;
var
  cp: LONGLONG;
begin
  if FLoaded then
  begin
    FSeeking.GetCurrentPosition(cp);
    if (cp = FDuration) and (FState = vpStop) then
      Position := 0;
    FControl.Run;
    FState := vpPlay;
    StartTimer;
  end;
end;

procedure TVideoPlayer.Stop;
begin
  if FLoaded then
  begin
    StopTimer;
    FControl.Stop;
    Position := 0;
    FControl.StopWhenReady;
    FState := vpStop;
  end;
end;

procedure TVideoPlayer.PauseOnly;
begin
  if FLoaded then
  begin
    StopTimer;
    FControl.Pause;
    FState := vpPause;
  end;
end;

procedure TVideoPlayer.Pause;
begin
  if FState = vpPause then
    Play
  else
    PauseOnly;
end;

function TVideoPlayer.SearchChunk(var F: File; Id: PChar; var Size: Integer): Boolean;
var
  sid: FOURCC;
  buf: array[0..3] of Char;
begin
  Result := False;
  try
    while not Eof(F) do
    begin
      BlockRead(F, sid, SizeOf(sid));
      BlockRead(F, Size, SizeOf(Size));
      if (sid = mmioStringToFOURCC('RIFF', 0)) or (sid = mmioStringToFOURCC('LIST', 0)) then
      begin
        BlockRead(F, buf, SizeOf(buf));
        if sid = mmioStringToFOURCC(Id, 0) then
        begin
          Result := True;
          Exit;
        end;
      end
      else begin
        if Id = 'RIFF' then
          Exit;
        if Size mod 2 = 1 then Inc(Size);
        if sid = mmioStringToFOURCC(Id, 0) then
        begin
          Result := True;
          Exit;
        end else
          Seek(F, FilePos(F)+Size);
      end;
    end;
  except
  end;
end;

{$DEFINE USEMMIO}
function TVideoPlayer.GetAVIHeader(const FileName: string): Boolean;
{$IFDEF USEMMIO}
var
  mmio: HMMIO;
  avi, hdrl, strl, strh, strf, avih: TMMCKINFO;
  i: Integer;
  ash: TAVIStreamHeader;
begin
  Result := False;
  with FFileInfo do
  begin
    FillChar(AVIMainHeader, SizeOf(AVIMainHeader), 0);
    FillChar(BitmapInfoHeader, SizeOf(BitmapInfoHeader), 0);
    FillChar(WaveFormat, SizeOf(WaveFormat), 0);
    avi.fccType := mmioStringToFOURCC('AVI ', 0);
    hdrl.fccType := mmioStringToFOURCC('hdrl', 0);
    avih.ckid := mmioStringToFOURCC('avih', 0);
    strl.fccType := mmioStringToFOURCC('strl', 0);
    strh.ckid := mmioStringToFOURCC('strh', 0);
    strf.ckid := mmioStringToFOURCC('strf', 0);

    mmio := mmioOpen(PChar(FileName), nil, MMIO_READ);
    if mmio = 0 then
      Exit;

    if mmioDescend(mmio, @avi, nil, MMIO_FINDRIFF) <> 0 then
      Exit;

    try
      if mmioDescend(mmio, @hdrl, @avi, MMIO_FINDLIST) <> 0 then
        Exit;

      if mmioDescend(mmio, @avih, @hdrl, MMIO_FINDCHUNK) <> 0 then
        Exit;
      mmioSeek(mmio, -8, SEEK_CUR);
      mmioRead(mmio, PChar(@AVIMainHeader), SizeOf(AVIMainHeader));
      mmioAscend(mmio, @avih, 0);

      for i := 0 to AVIMainHeader.dwStreams-1 do
      begin
        if mmioDescend(mmio, @strl, @hdrl, MMIO_FINDLIST) <> 0 then
          Exit;

        if mmioDescend(mmio, @strh, @strl, MMIO_FINDCHUNK) <> 0 then
          Exit;
        mmioSeek(mmio, -8, SEEK_CUR);
        mmioRead(mmio, PChar(@ash), SizeOf(ash));
        mmioAscend(mmio, @strh, 0);

        if mmioDescend(mmio, @strf, @strl, MMIO_FINDCHUNK) <> 0 then
          Exit;
        if ash.fccType =  mmioStringToFOURCC('vids', 0) then
          mmioRead(mmio, PChar(@BitmapInfoHeader), SizeOf(BitmapInfoHeader))
        else if ash.fccType =  mmioStringToFOURCC('auds', 0) then
          mmioRead(mmio, PChar(@WaveFormat), SizeOf(WaveFormat));
        mmioAscend(mmio, @strf, 0);

        mmioAscend(mmio, @strl, 0);
      end;
      Result := True;
    finally
      mmioClose(mmio, 0);
    end;
  end;
end;
{$ELSE}
var
  i, len, size: Integer;
  strh: TAVIStreamHeader;
  F: File;
begin
  Result := False;

  FileMode := 0;
  AssignFile(F, FileName);
  {$I-}
  Reset(F, 1);
  {$I+}
  if IOResult <> 0 then
    Exit;

  with FFileInfo do
  begin
    FillChar(AVIMainHeader, SizeOf(AVIMainHeader), 0);
    FillChar(BitmapInfoHeader, SizeOf(BitmapInfoHeader), 0);
    FillChar(WaveFormat, SizeOf(WaveFormat), 0);
    try
      if SearchChunk(F, 'RIFF', size) then
      begin
        SearchChunk(F, 'avih', size);
        Seek(F, FilePos(F)-8);
        BlockRead(F, AVIMainHeader, SizeOf(AVIMainHeader), len);
        for i := 0 to AVIMainHeader.dwStreams-1 do
        begin
          SearchChunk(F, 'strh', size);
          Seek(F, FilePos(F)-8);
          BlockRead(F, strh, SizeOf(strh), len);
          Seek(F, FilePos(F)-len+size+8);
          SearchChunk(F, 'strf', size);
          if strh.fccType =  mmioStringToFOURCC('vids', 0) then
            BlockRead(F, BitmapInfoHeader, SizeOf(BitmapInfoHeader), len)
          else if strh.fccType =  mmioStringToFOURCC('auds', 0) then
            BlockRead(F, WaveFormat, SizeOf(WaveFormat), len);
          Seek(F, FilePos(F)-len+size);
        end;
        Result := True;
      end ;
    except
    end;
  end;
  {
  if GetAVIHeader(F) then
    Aspect := 1
  else
    GetMpegInfo(F, Aspect, mpegw, mpegh);
  }
  CloseFile(F);
end;
{$ENDIF}

{
procedure TVideoPlayer.GetMpegInfo(var F: File; var Aspect: Double; var w, h: Integer);
const
  bufsize = 8192;
var
  buf: string;
  a, len, i: Integer;
begin
  a := 0;
  Seek(F, 0);
  SetLength(buf, bufsize);
  while True do
  begin
    FillChar(buf[1], bufsize, 0);
    BlockRead(F, buf[1], bufsize, len);
    i := Pos(#$00#$00#$01#$b3, buf);
    if (i > 0) and (len > (i+8)) then
    begin
      w := Byte(buf[i+4]) shl 4 + Byte(buf[i+5]) shr 4;
      h := Byte(buf[i+5]) and $F + Byte(buf[i+6]);
      a := Byte(buf[i+7]) shr 4;
      Break;
    end;
    // 1MBǂ񂾂炠߂
    if Eof(F) or (FilePos(F) > 1024*1024) then
      Break;
    Seek(F, FilePos(F)-4);
  end;
  if (a > 14) or (a < 0) then a := 0;
  Aspect := Mpeg1AspectTable[a];
end;
}

procedure TVideoPlayer.GetFilterNames;
var
  ief: IEnumFilters;
  Filter: IBaseFilter;
  FilterInfo: TFilterInfo;
  f: ULONG;
  prop: ISpecifyPropertyPages;
begin
  FilterCount := 0;
  FBuilder.EnumFilters(ief);
  while ief.Next(1, Filter, f) = S_OK do
  begin
    Filter.QueryFilterInfo(FilterInfo);
    Filter.QueryInterface(IID_ISpecifyPropertyPages, prop);
    if prop = nil then // vpeBȂtB^͖
      Continue;
    FilterNames[FilterCount] := FilterInfo.achName;
    Inc(FilterCount);
  end;
end;

function TVideoPlayer.LoadFile(const FileName: string): HResult;
var
  tpf: Double;
  //mpegw, mpegh: Integer;
  x, y: Integer;
begin
  Stop;
  FLoaded := False;
  FFileInfo.Length := 0;
  FFileInfo.FramePerSec := 0;
  FFileInfo.Aspect := 0;

  if Assigned(FControl) then
    FreeInterface;

  GetAVIHeader(FileName);

  Result := CoCreateInstance(CLSID_FilterGraph, nil, CLSCTX_INPROC_SERVER,
    IID_IGraphBuilder, FBuilder);
  try
    Result := FBuilder.RenderFile(PWideChar(WideString(FileName)), nil);
    if Result <> S_OK then
    begin
      FreeInterface;
      Exit;
    end;

    FWindow := FBuilder as IVideoWindow;
    FWindow.put_Owner(OAHWND(FParent));
    FWindow.put_WindowStyle(WS_CHILD);// or WS_CLIPSIBLINGS);
    FWindow.put_MessageDrain(FParent);
    FControl := FBuilder as IMediaControl;
    FSeeking := FBuilder as IMediaSeeking;
    FEvent := FBuilder as IMediaEvent;
    FSeeking.SetTimeFormat(TIME_FORMAT_MEDIA_TIME);
    FBasicVideo := FBuilder as IBasicVideo;
    FBasicVideo.GetVideoSize(FFileInfo.Width, FFileInfo.Height);
    FBasicVideo2 := FBuilder as IBasicVideo2;
    // MPEGǂ̔
    //if (Width <> mpegw) or (Height <> mpegh) then
    //  Aspect := 1;

    with FFileInfo do
    begin
      x := 0;
      y := 0;
      if FBasicVideo2.GetPreferredAspectRatio(x, y) = S_OK then
      begin
        if (x <> 0) and (y <> 0) then
        begin
          Aspect := (Width * y) / (Height * x);
          if Aspect= 1.125 then
            Aspect := 1.1; // NTSC
        end;
      end;

      tpf := 0;
      FBasicVideo.get_AvgTimePerFrame(tpf);
      if tpf = 0 then
      begin
        if AVIMainHeader.dwMicroSecPerFrame > 0 then
          FramePerSec := 1000000 / AVIMainHeader.dwMicroSecPerFrame
        else
          FramePerSec := 0;
      end else
        FramePerSec := 1 / tpf;

      FSeeking.GetDuration(FDuration);
      Length := Trunc(FDuration / 10000);
    end;

    FFileName := FileName;
    FLoaded := True;
    //Fivf := Figb as IVideoFrameStep;
    //FSeeking.SetRate(0.5);
    FAudio := FBuilder as IBasicAudio;

    GetFilterNames;
  except
    FreeInterface;
  end;
end;

procedure TVideoPlayer.SetWindowPosition(ALeft, ATop, AWidth, AHeight: Longint);
begin
  FWindow.SetWindowPosition(ALeft, ATop, AWidth, AHeight);
end;

function TVideoPlayer.WriteBitmapBits(const FileName: string): HRESULT;
var
  i, n: Integer;
  imd: IMediaDet;
  mt: TGUID;
begin
  if FLoaded then
  begin
    Result := CoCreateInstance(CLSID_MediaDet, nil, CLSCTX_INPROC_SERVER, IID_IMediaDet, imd);
    if Result <> S_OK then
      Exit;
    Result := imd.put_Filename(WideString(FFileName));
    if Result <> S_OK then
      Exit;
    Result := imd.get_OutputStreams(n);
    if Result <> S_OK then
      Exit;
    for i := 0 to n - 1 do
    begin
      Result := imd.put_CurrentStream(i);
      if Result <> S_OK then
        Continue;
      Result := imd.get_StreamType(mt);
      if Result <> S_OK then
        Continue;
      if CompareTGUID(mt, MEDIATYPE_Video) then
      begin
        Result := imd.WriteBitmapBits(Position / 1000.0, FFileInfo.Width, FFileInfo.Height, FileName);
        Exit;
      end;
    end;
  end;
  Result := E_FAIL;
end;
{
var
  GrabberF: IBaseFilter;
  Grabber: ISampleGrabber;
  amt: TAM_MEDIA_TYPE;
begin
  if FLoaded then
  begin
    Result := CoCreateInstance(CLSID_SampleGrabber, NULL, CLSCTX_INPROC_SERVER,
    IID_IBaseFilter, GrabberF);
    if Result <> S_OK then
      Exit;
    Result := Figb.AddFilter(GrabberF, 'Sample Grabber');
    if Result <> S_OK then
      Exit;
    Grabber := GrabberF as ISampleGrabber;
    FillChar(amt, SizeOf(amt), 0);
    amt.majortype := MEDIATYPE_Video;
    amt.subtype := MEDIATYPE_RGB24;
    amt.pbFormat
}

function TVideoPlayer.ShowFilterPropertyByName(const FilterName: string): HRESULT;
var
  Filter: IBaseFilter;
  prop: ISpecifyPropertyPages;
  pages: TCAGUID;
  FilterInfo: TFilterInfo;
  ief: IEnumFilters;
  f: ULONG;
begin
  Result := E_FAIL;
  if FLoaded then
  begin
    // FindFilterByName͂ȂWindows Media Source FilerŎs
    //{$C+}
    //if Figb.FindFilterByName(PWideChar(ws), Filter) <> S_OK then
    //  Assert(False);
    FBuilder.EnumFilters(ief);
    while ief.Next(1, Filter, f) = S_OK do
    begin
      Filter.QueryFilterInfo(FilterInfo);
      if FilterInfo.achName = FilterName then
      begin;
        Filter.QueryInterface(IID_ISpecifyPropertyPages, prop);
        if Assigned(prop) then
        begin
          Result :=  prop.GetPages(pages);
          if Result <> S_OK then
            Exit;
          OleCreatePropertyFrame(
            FParent,             // eEBhE
            0,                   // x (\ς)
            0,                   // y (\ς)
            FilterInfo.achName,  // _CAO {bNX̃LvV
            1,                   // tB^̐
            @Filter,             // tB^ւ̃|C^
            pages.cElems,        // vpeB y[W̐
            pages.pElems,        // vpeB y[W CLSID ւ̃|C^
            0,                   // P[ʎq
            0,                   // \ς
            nil                  // \ς
          );
          CoTaskMemFree(pages.pElems);
        end;
        Break;
      end;
    end;
  end;
end;

end.
