(*
 * Copyright (C) 2003-2004 Nobuyuki Honda (Ferdia)
 *
 * This software is provided under the zlib/libpng license.
 *)

program NkVp;

uses
  Windows, Messages, ShellAPI, ActiveX,
  NkdtProf, VideoPlayer, SUtil, SForm, SCmd, SConfFile, SList,
  Command;

{$R *.res}

const
  AppName = 'NkVp';
  Version = '0.5.3 (2004-07-03)';
  Copyright = 'Copyright (C) 2003-2004 Nobuyuki Honda (Ferdia)';
  License = 'This software is provided under the zlib/libpng license.';

  WM_USER_LOADFILE = WM_APP + 1;

type
  TFileList = class(TSStrList)
    CurPos: Integer;
    procedure Clear;
    function CurFileName: string;
  end;

  TAspcetMode = (amAuto, amOrg, am43, am169);

  TNkVpMainForm = class(TSForm)
  private
    FHideCursor: Boolean;
    procedure SetHideCursor(Value: Boolean);
  public
    StateStr: string;
    TimeStr: string;
    FormatStr: string;
    AspectMode: TAspcetMode;
    Zoom: Integer;
    DispWidth: Integer;
    DispHeight: Integer;
    Mute: Boolean;
    SaveState: TVPState;
    RepeatFlg: Boolean;
    HideCursorCount: Integer;
    OldPt: TPoint;
    procedure InitParams; override;
    procedure DoCmd(const CmdName: string);
    procedure SetTitle;
    procedure SetFormatStr;
    procedure FitWindow;
    procedure ZoomVideo(Value: Integer);
    procedure ShowAbout;
    procedure PopupMenu(const pt: TPoint);
    procedure PopupFileListMenu(const pt: TPoint);
    procedure DrawSeekBar(dc: HDC);
    procedure OnTimer;
    procedure OnEndPlay;
    procedure SaveScreenCapture;
    procedure LoadFile(const FileName: string);
    function GetVideoDispWidth: Integer;
    procedure UpdateVideoWindowSize;
    function GetSeekBarRect: TRect;

    procedure WMMouseWheel(var Message: TMessage); message WM_MOUSEWHEEL;
    procedure WMTimer(var Message: TWMTimer); message WM_TIMER;
    procedure WMGetMinMaxInfo(var Message: TWMGetMinMaxInfo); message WM_GETMINMAXINFO;
    procedure WMDropFiles(var Message: TWMDropFiles); message WM_DROPFILES;
    procedure WMCommand(var Message: TWMCommand); message WM_COMMAND;
    procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
    procedure WMContextMenu(var Message: TMessage); message WM_CONTEXTMENU;
    procedure WMMButtonUp(var Message: TWMMButtonUp); message WM_MBUTTONUP;
    procedure WMMouseMove(var Message: TWMMouseMove); message WM_MOUSEMOVE;
    procedure WMLButtonDown(var Message: TWMLButtonDown); message WM_LBUTTONDOWN;
    procedure WMLButtonDblClk(var Message: TWMLButtonDblClk); message WM_LBUTTONDBLCLK;
    procedure WMLButtonUp(var Message: TWMLButtonUp); message WM_LBUTTONUP;
    procedure WMMove(var Message: TWMMove); message WM_MOVE;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure WMCopyData(var Message: TWMCopyData); message WM_COPYDATA;
    procedure WMDestroy(var Message: TWMDestroy); message WM_DESTROY;
    procedure WMUserLoadFile(var Message); message WM_USER_LOADFILE;

    property HideCursor: Boolean read FHideCursor write SetHideCursor;
  end;

  TWaveFmtTable = record
    Tag: Integer;
    Name: string;
  end;

const
  WaveFormatTable: array[0..2] of TWaveFmtTable = (
    (Tag: 1;  Name: 'PCM'),
    (Tag: 2;  Name: 'MSADPCM'),
    (Tag: 85; Name: 'MP3')
  );

  IDM_FILTERPROP_FIRST = 4000;
  IDM_FILTERPROP_LAST = IDM_FILTERPROP_FIRST + 255;
  IDM_FILELIST_FIRST = 5000;
  IDM_FILELIST_LAST = IDM_FILELIST_FIRST + 10000;

  KeyMapFile = 'keymap';
  CaptureBmpFile = 'nkvpcapt.bmp';

  MsgSaveScreenCaptureError = 'Cannot save screen capture.%n%n%1%n%n';
  MsgCannotPlay = 'Cannot play.%n%n%1%n%n';

  SeekBarHeight = 13;
  NormalSeek = 1000; // msec
  FastSeek = 10000;
  FasterSeek = 60000;
  HideCursorDelay = 3; // }EXJ[\܂ł̎(sec)

var
  MainForm: TNkVpMainForm;
  Video: TVideoPlayer;
  Cmd: TSCmd;
  ProfDir: string;
  FileList: TFileList;
  ExitWhenPlayEnd: Boolean;

{ TFileList }

procedure TFileList.Clear;
begin
  CurPos := 0;
  inherited;
end;

function TFileList.CurFileName: string;
begin
  Result := Strings[CurPos];
end;

{ TNkVpMainForm }

procedure TNkVpMainForm.InitParams;
begin
  inherited;
  ClassStyle := CS_HREDRAW or CS_VREDRAW or CS_DBLCLKS;
  Style := WS_OVERLAPPEDWINDOW or WS_CLIPCHILDREN;
  ExStyle := WS_EX_ACCEPTFILES;
  // NULL_BRUSHFullScreen؂ւVideoWindowĕ`悳Ȃ
  Brush := GetStockObject(BLACK_BRUSH);
end;

procedure TNkVpMainForm.ShowAbout;
var
  s: string;
begin
  if Video.FileInfo.Aspect = 0 then
    s := 'not available'
  else
    s := FStr(Video.FileInfo.Aspect, 4);
  AboutBox(Handle, 'About '+AppName, SStr([
    AppName+' '+Version+#13#10,
    #13#10,
    Copyright+#13#10,
    #13#10,
    License+#13#10,
    #13#10,
    'Profile Folder: '+ProfDir+#13#10,
    #13#10,
    'Window Size: ', DispWidth, 'x', DispHeight, #13#10,
    'File Aspcet: '+s])
  );
end;

function TNkVpMainForm.GetVideoDispWidth: Integer;
begin
  with Video do
    case AspectMode of
      amOrg:
        Result := FileInfo.Width;
      amAuto:
        if FileInfo.Aspect > 0 then
          Result := Trunc(FileInfo.Width / FileInfo.Aspect + 0.5)
        else
          Result := FileInfo.Width;
      am43:
        Result := Trunc(FileInfo.Height * 4/3 + 0.5);
      am169:
        Result := Trunc(FileInfo.Height * 16/9 + 0.5);
    else
      Result := FileInfo.Width;
    end;
end;

procedure TNkVpMainForm.FitWindow;
var
  w, h: Integer;
begin
  if Video.Loaded then
  begin
    Restore;
    w := GetVideoDispWidth * Zoom div 100;
    h := Video.FileInfo.Height * Zoom div 100 + SeekBarHeight;
    ClientRect := Rect(0, 0, w, h);
    Video.SetWindowPosition(0, 0, w, h-SeekBarHeight);
  end;
end;

procedure TNkVpMainForm.SetFormatStr;
  function GetWaveFormatName(Tag: Integer): string;
  var
    i: Integer;
  begin
    for i := 0 to High(WaveFormatTable) do
      if WaveFormatTable[i].Tag = Tag then
      begin
        Result := WaveFormatTable[i].Name + ' ';
        Exit;
      end;
    Result := '';
  end;
begin
  FormatStr := '';
  if Video.Loaded then
    with Video.FileInfo do
    begin
      FormatStr := GetWaveFormatName(WaveFormat.wFormatTag);
      if WaveFormat.nSamplesPerSec <> 0 then
        FormatStr := FormatStr + SStr([WaveFormat.nSamplesPerSec div 1000, 'kHz ',
          WaveFormat.nAvgBytesPerSec*8 div 1000, 'k ']);
      if WaveFormat.nChannels = 2 then
        FormatStr := FormatStr + 'st ';
      FormatStr := FormatStr + Copy(PChar(@BitmapInfoHeader.biCompression), 1, 4) + ' ';
    end;
end;

procedure TNkVpMainForm.UpdateVideoWindowSize;
var
  cw, ch, vw, vh: Integer;
begin
  if Video.Loaded then
  begin
    cw := ClientRect.Right;
    ch := ClientRect.Bottom - SeekBarHeight;
    vw := GetVideoDispWidth;
    vh := Video.FileInfo.Height;
    if vh = 0 then
      Exit;
    if vw*ch > vh*cw then
    begin
      DispWidth := cw;
      DispHeight := (vh * DispWidth) div vw;
    end
    else begin
      DispHeight := ch;
      DispWidth := (vw * DispHeight) div vh;
    end;
    if DispHeight > DispWidth then
      Zoom := (DispWidth * 100) div vw
    else
      Zoom := (DispHeight * 100) div vh;
    Video.SetWindowPosition((cw-DispWidth) div 2,
      (ch-DispHeight) div 2, DispWidth, DispHeight);
    SetTitle;
  end;
end;

procedure TNkVpMainForm.SetTitle;
begin
  with Video do
    if Loaded then
      Text := SStr([BaseName(FileName)+' (', FileInfo.Width, 'x', FileInfo.Height,
        ' '+FStr(FileInfo.FramePerSec, 3)+'fps) ', Zoom, '% - '+AppName])
    else
      Text := AppName;
end;

procedure TNkVpMainForm.SaveScreenCapture;
var
  SaveFileName: string;
  r: HRESULT;
begin
  Video.PauseOnly;
  OnTimer;
  Update; // ĕ`悵ĂȂƍcۂ
  with TSSaveDlg.Create(Handle, ['All Files', '*.*'], CaptureBmpFile) do
    try
      if ShowModal(SaveFileName) then
      begin
        r := Video.WriteBitmapBits(SaveFileName);
        //r := E_UNEXPECTED; // test
        if r <> S_OK then
          FormatMsgBox(Handle, AppName, MsgSaveScreenCaptureError+GetDShowErrMsg(r),
            [SaveFileName], MB_ICONEXCLAMATION);
      end;
    finally
      Free;
    end;
end;

procedure TNkVpMainForm.OnEndPlay;
begin
  if RepeatFlg then
  begin
    Video.Position := 0;
    Video.Play;
  end
  else if FileList.CurPos < FileList.Count - 1 then
  begin
    Inc(FileList.CurPos);
    PostMessage(Handle, WM_USER_LOADFILE, 0, 0); //ʃXbhȂ̂PostMessae
  end
  else if ExitWhenPlayEnd then
    PostMessage(Handle, WM_SYSCOMMAND, SC_CLOSE, 0); //ʃXbhȂ̂PostMessae
end;

procedure TNkVpMainForm.OnTimer;
  function GetTimeStr(n: Integer): string;
  var
    t: Integer;
  begin
    t := n div 1000;
    Result := IZStr(t div 3600, 2) + ':' +
              IZStr(t mod 3600 div 60, 2) + ':' +
              IZStr(t mod 60, 2);
  end;
var
  rc: TRect;
begin
  TimeStr := FormatStr + GetTimeStr(Video.Position) + ' / ' + GetTimeStr(Video.FileInfo.Length) + ' ';
  StateStr := '';
  if FileList.Count > 1 then
    StateStr := SStr([' ', FileList.CurPos+1, '/', FileList.Count, ':']);
  case Video.State of
    vpPause: StateStr := StateStr + ' PAUSE';
    vpStop:  StateStr := StateStr + ' STOP';
  end;
  if RepeatFlg then
    StateStr := StateStr + ' REPEAT';
  if Mute then
    StateStr := StateStr + ' MUTE';
  rc := GetSeekBarRect;
  InvalidateRect(Handle, @rc, False);
end;

function TNkVpMainForm.GetSeekBarRect: TRect;
begin
  Result := ClientRect;
  Result.Top := Result.Bottom - SeekBarHeight;
end;

procedure TNkVpMainForm.DrawSeekBar(dc: HDC);
var
  src, rc, rc2: TRect;
  bbr, gbr: HBRUSH;
  font: HFONT;
  mdc: HDC;
  bmp, oldbmp: HBITMAP;
begin
  rc := GetSeekBarRect;
  src.Left := 0;
  src.Top := 0;
  src.Right :=  rc.Right - rc.Left;
  src.Bottom := rc.Bottom - rc.Top;
  mdc := CreateCompatibleDC(dc);
  bmp := CreateCompatibleBitmap(dc, src.Right, src.Bottom);
  oldbmp := SelectObject(mdc, bmp);

  bbr := GetStockObject(BLACK_BRUSH);
  gbr := GetStockObject(GRAY_BRUSH);
  FrameRect(mdc, src, gbr);
  Inc(src.Top);
  FillRect(mdc, src, bbr);
  if Video.Loaded and (Video.FileInfo.Length > 0) then
    rc2.Left := Trunc(Video.Position / Video.FileInfo.Length * rc.Right) - 2
  else
    rc2.Left := -2;
  rc2.Right := rc2.Left + 4 + 1;
  rc2.Top := 2;
  rc2.Bottom := src.Bottom - 1;
  FillRect(mdc, rc2, gbr);
  Inc(rc2.Left ,2);
  rc2.Right := rc2.Left + 1;
  rc2.Bottom := rc2.Top + 2;
  FillRect(mdc, rc2, bbr);
  rc2.Bottom := src.Bottom - 1;
  rc2.Top := rc2.Bottom - 2;
  FillRect(mdc, rc2, bbr);
  SetTextColor(mdc, RGB(0, 255, 255));
  SetBkMode(mdc, TRANSPARENT);
  font := SelectObject(mdc, GetStockObject(DEFAULT_GUI_FONT));
  DrawText(mdc, PChar(StateStr), Length(StateStr), src, DT_SINGLELINE or DT_VCENTER);
  DrawText(mdc, PChar(TimeStr), Length(TimeStr), src, DT_RIGHT or DT_SINGLELINE or DT_VCENTER);
  SelectObject(mdc, font);
  BitBlt(dc, rc.Left, rc.Top, rc.Right, rc.Bottom, mdc, 0, 0, SRCCOPY);
  SelectObject(mdc, oldbmp);
  DeleteObject(bmp);
  DeleteDC(mdc);
end;

procedure TNkVpMainForm.ZoomVideo(Value: Integer);
begin
  if FullScreen then
  begin
    FullScreen := False;
    HideCursor := False;
    KillTimer(Handle, 1);
  end;
  Zoom := Value;
  FitWindow;
  SetTitle;
end;

procedure TNkVpMainForm.SetHideCursor(Value: Boolean);
begin
  if Value <> FHideCursor then
    ShowCursor(not Value);
  FHideCursor := Value;
  if not HideCursor then
    HideCursorCount := 0;
end;

procedure TNkVpMainForm.DoCmd(const CmdName: string);
begin
  if CmdName = '' then
    Exit
  else if CmdName = cmPause then
  begin
    if Video.State = vpStop then
      Video.Play
    else
      Video.Pause;
  end
  else if CmdName = cmStop then
    Video.Stop
  else if CmdName = cmSkipPrev then
  begin
    if FileList.CurPos > 0 then
    begin
      Dec(FileList.CurPos);
      LoadFile(FileList.CurFileName);
    end;
  end
  else if CmdName = cmSkipNext then
  begin
    if FileList.CurPos < FileList.Count - 1 then
    begin
      Inc(FileList.CurPos);
      LoadFile(FileList.CurFileName);
    end;
  end
  else if Copy(CmdName, 1, 4) = Copy(cmZoom100, 1, 4) then
    ZoomVideo(SInt(Copy(CmdName, 5, Length(CmdName)-5+1)))
  else if CmdName = cmStayOnTop then
    StayOnTop := not StayOnTop
  else if CmdName = cmFullScreen then
  begin
    FullScreen := not FullScreen;
    if FullScreen then
      SetTimer(Handle, 1, 500, nil)
    else begin
      HideCursor := False;
      KillTimer(Handle, 1);
    end;
  end
  else if CmdName = cmMute then
  begin
    Mute := not Mute;
    if Mute then
      Video.Volume := -10000
    else
      Video.Volume := 0;
  end
  else if CmdName = cmSeekForward then
    Video.Position := Video.Position + NormalSeek
  else if CmdName = cmSeekBack then
    Video.Position := Video.Position - NormalSeek
  else if CmdName = cmSeekForwardFast then
    Video.Position := Video.Position + FastSeek
  else if CmdName = cmSeekBackFast then
    Video.Position := Video.Position - FastSeek
  else if CmdName = cmSeekForwardFaster then
    Video.Position := Video.Position + FasterSeek
  else if CmdName = cmSeekBackFaster then
    Video.Position := Video.Position - FasterSeek
  else if CmdName = cmAspectAuto then
  begin
    AspectMode := amAuto;
    UpdateVideoWindowSize;
  end
  else if CmdName = cmAspectOrg then
  begin
    AspectMode := amOrg;
    UpdateVideoWindowSize;
  end
  else if CmdName = cmAspect43 then
  begin
    AspectMode := am43;
    UpdateVideoWindowSize;
  end
  else if CmdName = cmAspect169 then
  begin
    AspectMode := am169;
    UpdateVideoWindowSize;
  end
  else if CmdName = cmScreenCapture then
    SaveScreenCapture
  else if CmdName = cmRepeat then
    RepeatFlg := not RepeatFlg
  else if CmdName = cmAbout then
    ShowAbout
  else if CmdName = cmExit then
    Close;
  OnTimer;
end;

procedure TNkVpMainForm.PopupMenu(const pt: TPoint);
var
  Menu, SubMenu: HMENU;
  i: Integer;
begin
  with TSCmdMenu.Create(Cmd, MenuTable) do
    try
      if not Video.Loaded then
        DisableItems([
          cmPause,
          cmStop,
          cmFullScreen,
          cmZoom50,
          cmZoom100,
          cmZoom150,
          cmZoom200,
          cmZoom300,
          cmZoom400,
          cmScreenCapture,
          '{FilterMenu'
        ]);

      CheckItem(cmFullScreen, FullScreen);
      CheckItem(cmStayOnTop, StayOnTop);
      CheckItem(cmMute, Mute);
      CheckItem(cmRepeat, RepeatFlg);
      CheckItem(cmAspectAuto, AspectMode = amAuto);
      CheckItem(cmAspectOrg, AspectMode = amOrg);
      CheckItem(cmAspect43, AspectMode = am43);
      CheckItem(cmAspect169, AspectMode = am169);

      Menu := MakeMenuHandle;
    finally
      Free;
    end;

  SubMenu := GetSubMenu(Menu, GetMenuItemCount(Menu) - 5);
  for i := 0 to Video.FilterCount - 1 do
    AppendMenu(SubMenu, MF_STRING, IDM_FILTERPROP_FIRST+i, PChar(Video.FilterNames[i]));

  TrackPopupMenu(Menu, TPM_LEFTALIGN or TPM_LEFTBUTTON or TPM_RIGHTBUTTON,
    pt.x, pt.y, 0, Handle, nil);

  DestroyMenu(Menu);
end;

procedure TNkVpMainForm.PopupFileListMenu(const pt: TPoint);
var
  Menu: HMENU;
  i: Integer;
  s: string;
begin
  Menu := CreatePopupMenu;
  for i := 0 to FileList.Count - 1 do
  begin
    s := IStr(i+1) + ' ' + BaseName(FileList[i]);
    if i < 9 then
      s := '0&' + s;
    AppendMenu(Menu, MF_STRING, IDM_FILELIST_FIRST+i, PChar(s));
  end;
  CheckMenuItem(Menu, FileList.CurPos, MF_BYPOSITION or MF_CHECKED);

  TrackPopupMenu(Menu, TPM_LEFTALIGN or TPM_LEFTBUTTON or TPM_RIGHTBUTTON,
    pt.x, pt.y, 0, Handle, nil);

  DestroyMenu(Menu);
end;

procedure TNkVpMainForm.WMMouseWheel(var Message: TMessage);
begin
  with Message do
    if WParam > 0 then
    begin
      if GetKeyState(VK_SHIFT) < 0 then
        DoCmd(cmSeekBackFast)
      else
        DoCmd(cmSeekBack);
    end
    else begin
      if GetKeyState(VK_SHIFT) < 0 then
        DoCmd(cmSeekForwardFast)
      else
        DoCmd(cmSeekForward);
    end;
end;

procedure TNkVpMainForm.WMTimer(var Message: TWMTimer);
var
  pt: TPoint;
begin
  GetCursorPos(pt);
  if (OldPt.x = pt.x) and (OldPt.y = pt.y) then
  begin
    if HideCursorCount = HideCursorDelay*2 then
      HideCursor := True;
    Inc(HideCursorCount);
  end else
    HideCursor := False;
  OldPt := pt;
end;

procedure TNkVpMainForm.WMGetMinMaxInfo(var Message: TWMGetMinMaxInfo);
begin
  with Message do
  begin
    MinMaxInfo.ptMaxTrackSize.x := 4096;
    MinMaxInfo.ptMaxTrackSize.y := 4096;
  end;
end;

procedure TNkVpMainForm.WMDropFiles(var Message: TWMDropFiles);
var
  FileName: string;
  cnt, i: Integer;
begin
  FileList.Clear;
  with Message do
  begin
    cnt := DragQueryFile(Drop, $FFFFFFFF, nil, 0);
    for i := 0 to cnt - 1 do
    begin
      SetLength(FileName, DragQueryFile(Drop, i, nil, 0)+1);
      DragQueryFile(Drop, i, PChar(FileName), Length(FileName));
      // null terminateȂ̂PCharɃLXg
      FileList.Add(PChar(FileName));
    end;
    DragFinish(Drop);
  end;
  FileList.Sort;
  LoadFile(FileList.CurFileName);
  Zoom := 100;
  FitWindow;
end;

procedure TNkVpMainForm.WMCommand(var Message: TWMCommand);
var
  r: HRESULT;
begin
  with Message do
    case ItemID of
      IDM_FILTERPROP_FIRST .. IDM_FILTERPROP_LAST:
        begin
          r := Video.ShowFilterPropertyByName(Video.FilterNames[ItemID-IDM_FILTERPROP_FIRST]);
          if r <> 0 then
            FormatMsgBox(Handle, AppName, '%1%n%n'+GetDShowErrMsg(r),
              [Video.FilterNames[ItemID-IDM_FILTERPROP_FIRST]], MB_ICONEXCLAMATION);
        end;
      IDM_FILELIST_FIRST .. IDM_FILELIST_LAST:
        begin
          FileList.CurPos := ItemID-IDM_FILELIST_FIRST;
          LoadFile(FileList.CurFileName);
        end;
    else
      DoCmd(Cmd.Name(ItemID));
    end;
end;

procedure TNkVpMainForm.WMPaint(var Message: TWMPaint);
var
  dc: HDC;
  ps: TPaintStruct;
  rc: TRect;
begin
  dc := BeginPaint(Handle, ps);
  rc := ClientRect;
  Dec(rc.Bottom, SeekBarHeight);
  //FillRect(dc, rc, GetStockObject(GRAY_BRUSH));
  DrawSeekBar(dc);
  EndPaint(Handle, ps);
end;

procedure TNkVpMainForm.WMContextMenu(var Message: TMessage);
var
  pt: TPoint;
begin
  with Message do
    if GetContextMenuPos(Handle, LParam, 0, 0, pt) then
    begin
      if PtInRect(GetSeekBarRect, ScreenToClient(pt)) then
        PopupFileListMenu(pt)
      else
        PopupMenu(pt)
    end else
      inherited;
end;

procedure TNkVpMainForm.WMMButtonUp(var Message: TWMMButtonUp);
begin
  DoCmd(cmStop);
end;

procedure TNkVpMainForm.WMMouseMove(var Message: TWMMouseMove);
var
  pt: TPoint;
begin
  if GetCapture = Handle then
  begin
    GetCursorPos(pt);
    pt := ScreenToClient(pt);
    Video.Position := Trunc(pt.x / ClientRect.Right * Video.FileInfo.Length);
    OnTimer;
  end;
end;

procedure TNkVpMainForm.WMLButtonDown(var Message: TWMLButtonDown);
var
  rc: TRect;
begin
  rc := GetSeekBarRect;
  with Message do
    if PtInRect(rc, Point(XPos, YPos)) then
    begin
      SetCapture(Handle);
      SaveState := Video.State;
      Video.PauseOnly;
      Video.Position := Trunc(XPos / rc.Right * Video.FileInfo.Length);
      OnTimer;
    end else
      DoCmd(cmPause)
end;

procedure TNkVpMainForm.WMLButtonDblClk(var Message: TWMLButtonDblClk);
begin
  WMLButtonDown(Message);
end;

procedure TNkVpMainForm.WMLButtonUp(var Message: TWMLButtonUp);
begin
  if GetCapture = Handle then
  begin
    ReleaseCapture;
    if SaveState = vpPlay then
      Video.Play;
  end;
end;

procedure TNkVpMainForm.WMMove(var Message: TWMMove);
begin
  if Video.Loaded then
    InvalidateRect(GetWindow(Handle, GW_CHILD), nil, False); // VideoWindowXV
end;

procedure TNkVpMainForm.WMSize(var Message: TWMSize);
begin
  UpdateVideoWindowSize;
end;

procedure TNkVpMainForm.WMCopyData(var Message: TWMCopyData);
var
  FileName: string;
begin
  Message.Result := Integer(True);
  with Message.CopyDataStruct^ do
  begin
    if dwData <> 0 then
      Exit;
    SetLength(FileName, cbData);
    Move(lpData^, FileName[1], cbData);
  end;

  Restore;

  FileList.Clear;
  FileList.Add(FileName);
  PostMessage(Handle, WM_USER_LOADFILE, 1, 0); //͕ʃXbh?Ȃ̂
end;

procedure TNkVpMainForm.WMDestroy(var Message: TWMDestroy);
begin
  Video.PauseOnly;
  Video.FreeInterface;
end;

procedure TNkVpMainForm.WMUserLoadFile(var Message);
begin
  with TMessage(Message) do
  begin
    LoadFile(FileList.CurFileName);
    if wParam = 1 then
    begin
      Zoom := 100;
      FitWindow;
    end;
  end;
end;

procedure TNkVpMainForm.LoadFile(const FileName: string);
var
  r: HResult;
begin
  r := Video.LoadFile(FileName);
  if r = S_OK then
  begin
    AspectMode := amAuto;
    SetTitle;
    SetFormatStr;
    if Mute then
      Video.Volume := -10000
    else
      Video.Volume := 0;
    Video.Play;
    UpdateVideoWindowSize; // ĐɃTCYXVȂƕ`悳Ȃ
  end
  else begin
    SetTitle;
    SetFormatStr;
    OnTimer;
    FormatMsgBox(Handle, AppName, MsgCannotPlay+GetDShowErrMsg(r),
      [FileName], MB_ICONEXCLAMATION);
  end;
end;

procedure LoadFileByOtherApp(TargetWnd: HWND; const FileName: string);
var
  cds: TCopyDataStruct;
begin
  cds.dwData := 0;
  cds.cbData := Length(FileName);
  cds.lpData := PChar(FileName);
  SendMessage(TargetWnd, WM_COPYDATA, 0, LPARAM(@cds));
end;

var
  Wnd: HWND;
  i: Integer;
  s: string;
begin
  FileList := TFileList.Create;
  for i := 1 to ParamCount do
  begin
    s := ParamStr(i);
    if s = '-e' then
      ExitWhenPlayEnd := True
    else
      FileList.Add(s);
  end;

  if AlreadyRun(AppName) then
  begin
    s := TNkVpMainForm.ClassName;
    Wnd := FindWindow(PChar(s), nil);
    if Wnd <> 0 then
    begin
      if FileList.Count > 0 then
      begin
        SetForegroundWindow(Wnd);
        LoadFileByOtherApp(Wnd, FileList.CurFileName);
        FileList.Free;
        Exit;
      end;
    end;
  end;

  CoInitialize(nil);
  try
    ProfDir := GetProfDir('nkvp') + '\';

    Cmd := TSCmd.Create(CmdList, DefaultKeys, ProfDir + KeyMapFile);
    MainForm := TNkVpMainForm.CreateDefaultPos(AppName, 320, 240);
    try
      if MainForm.Handle = 0 then
        Halt(255);

      SetMainForm(MainForm);

      Video := TVideoPlayer.Create(MainForm.Handle);
      try
        Video.OnTimer := MainForm.OnTimer;
        Video.OnEndPlay := MainForm.OnEndPlay;

        MainForm.Show;

        if FileList.Count > 0 then
        begin
          MainForm.LoadFile(FileList.CurFileName);
          MainForm.Zoom := 100;
          MainForm.FitWindow;
        end;
        MainForm.OnTimer;

        ExitCode := MessageLoopWithAccel(MainForm.Handle, Cmd.AccHandle);

      finally
        Video.Free;
      end;
    finally
      MainForm.Free;
      Cmd.Free;
      FileList.Free;
    end;
  finally
    CoUninitialize;
  end;
end.

